==> bbmap-35.85-2.tar.bz2: opt/bbmap-35.85/current/jgi/FungalRelease.java <==
package jgi;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;

import stream.ConcurrentReadInputStream;
import stream.FASTQ;
import stream.FastaReadInputStream;
import stream.ConcurrentReadOutputStream;
import stream.Read;

import dna.Parser;
import dna.Timer;
import fileIO.ByteFile;
import fileIO.ReadWrite;
import fileIO.FileFormat;
import fileIO.TextStreamWriter;

import align2.ListNum;
import align2.ReadLengthComparator;
import align2.ReadStats;
import align2.Shared;
import align2.Tools;

/**
 * Reformats a fungal assembly for release.
 * Also creates contig and agp files.
 * 
 * @author Brian Bushnell
 * @date December 9, 2015
 *
 */
public class FungalRelease {
	
	/*--------------------------------------------------------------*/
	/*----------------        Initialization        ----------------*/
	/*--------------------------------------------------------------*/
	
	/**
	 * Code entrance from the command line.
	 * @param args Command line arguments
	 */
	public static void main(String[] args){
		Timer t=new Timer();
		FungalRelease fr=new FungalRelease(args);
		fr.process(t);
	}
	
	/**
	 * Constructor.
	 * @param args Command line arguments
	 */
	public FungalRelease(String[] args){
		
		//Process any config files
		args=Parser.parseConfig(args);
		
		//Detect whether the uses needs help
		if(Parser.parseHelp(args, true)){
			printOptions();
			System.exit(0);
		}
		
		//Print the program name and arguments
		outstream.println("Executing "+getClass().getName()+" "+Arrays.toString(args)+"\n");
		
		FASTQ.FORCE_INTERLEAVED=FASTQ.TEST_INTERLEAVED=false;
		Shared.FASTA_WRAP=60;
		
		//Set some shared static variables regarding PIGZ
		Shared.READ_BUFFER_LENGTH=Tools.min(200, Shared.READ_BUFFER_LENGTH);
		Shared.capBuffers(4);
		ReadWrite.USE_PIGZ=ReadWrite.USE_UNPIGZ=true;
		ReadWrite.MAX_ZIP_THREADS=Shared.threads();
		
		Read.TO_UPPER_CASE=true;
		
		//Create a parser object
		Parser parser=new Parser();
		
		//Parse each argument
		for(int i=0; i<args.length; i++){
			String arg=args[i];
			
			//Break arguments into their constituent parts, in the form of "a=b"
			String[] split=arg.split("=");
			String a=split[0].toLowerCase();
			String b=split.length>1 ? split[1] : null;
			if(b==null || b.equalsIgnoreCase("null")){b=null;}
			while(a.startsWith("-")){a=a.substring(1);} //Strip leading hyphens
			
			
			if(parser.parse(arg, a, b)){//Parse standard flags in the parser
				//do nothing
			}else if(a.equals("verbose")){
				verbose=Tools.parseBoolean(b);
			}else if(a.equals("mingapin")){
				minGapIn=(int)Tools.parseKMG(b);
			}else if(a.equals("mingap") || a.equals("mingapout")){
				minGapOut=(int)Tools.parseKMG(b);
			}else if(a.equals("minlen") || a.equals("minlength") || a.equals("minscaf")){
				minScaf=(int)Tools.parseKMG(b);
			}else if(a.equals("mincontig")){
				minContig=(int)Tools.parseKMG(b);
			}else if(a.equals("outc") || a.equals("contigs")){
				outC=b;
			}else if(a.equals("qfoutc")){
				qfoutC=b;
			}else if(a.equals("sortcontigs")){
				sortContigs=Tools.parseBoolean(b);
			}else if(a.equals("sortcscaffolds")){
				sortScaffolds=Tools.parseBoolean(b);
			}else if(a.equals("baniupac")){
				banIupac=Tools.parseBoolean(b);
			}else if(a.equals("agp")){
				agpFile=b;
			}else if(a.equals("legend")){
				legendFile=b;
			}else if(a.equals("scafnum")){
				scafNum=Tools.parseKMG(b);
			}else if(a.equals("renamescaffolds") || a.equals("rename")){
				renameScaffolds=Tools.parseBoolean(b);
			}else if(a.equals("scafnum")){
				contigNum=Tools.parseKMG(b);
			}else if(a.equals("renamecontigs")){
				renameContigs=Tools.parseBoolean(b);
			}else if(a.equals("parse_flag_goes_here")){
				//Set a variable here
			}else{
				outstream.println("Unknown parameter "+args[i]);
				assert(false) : "Unknown parameter "+args[i];
				//				throw new RuntimeException("Unknown parameter "+args[i]);
			}
		}
		
		{//Process parser fields
			Parser.processQuality();
			
			maxReads=parser.maxReads;
			
			overwrite=ReadStats.overwrite=parser.overwrite;
			append=ReadStats.append=parser.append;
			
			in1=parser.in1;
			qfin1=parser.qfin1;
			
			out1=parser.out1;
			qfout1=parser.qfout1;
			
			extin=parser.extin;
			extout=parser.extout;
		}
		
		assert(FastaReadInputStream.settingsOK());
		
		//Ensure there is an input file
		if(in1==null){
			printOptions();
			throw new RuntimeException("Error - at least one input file is required.");
		}
		
		//Adjust the number of threads for input file reading
		if(!ByteFile.FORCE_MODE_BF1 && !ByteFile.FORCE_MODE_BF2 && Shared.threads()>2){
			ByteFile.FORCE_MODE_BF2=true;
		}
		
		//Ensure output files can be written
		if(!Tools.testOutputFiles(overwrite, append, false, out1, outC)){
			outstream.println((out1==null)+", "+out1);
			throw new RuntimeException("\n\noverwrite="+overwrite+"; Can't write to output files "+out1+"\n");
		}
		
		//Ensure input files can be read
		if(!Tools.testInputFiles(false, true, in1)){
			throw new RuntimeException("\nCan't read to some input files.\n");
		}
		
		//Ensure that no file was specified multiple times
		if(!Tools.testForDuplicateFiles(true, in1, out1, outC)){
			throw new RuntimeException("\nSome file names were specified multiple times.\n");
		}
		
		//Create output FileFormat objects
		ffout1=FileFormat.testOutput(out1, FileFormat.FASTA, extout, true, overwrite, append, ordered);
		
		//Create output FileFormat objects
		ffoutC=FileFormat.testOutput(outC, FileFormat.FASTA, extout, true, overwrite, append, ordered);

		//Create input FileFormat objects
		ffin1=FileFormat.testInput(in1, FileFormat.FASTA, extin, true, true);
	}
	
	/*--------------------------------------------------------------*/
	/*----------------         Outer Methods        ----------------*/
	/*--------------------------------------------------------------*/

	/** Create read streams and process all data */
	void process(Timer t){
		
		//Create a read input stream
		final ConcurrentReadInputStream cris;
		{
			cris=ConcurrentReadInputStream.getReadInputStream(maxReads, true, ffin1, null, qfin1, null);
			cris.start(); //Start the stream
			if(verbose){outstream.println("Started cris");}
		}
		
		//Optionally create a read output stream
		final ConcurrentReadOutputStream ros, rosc;
		final int buff=4;
		if(ffout1!=null){
			ros=ConcurrentReadOutputStream.getStream(ffout1, null, qfout1, null, buff, null, false);
			ros.start(); //Start the stream
		}else{ros=null;}
		if(ffoutC!=null){
			rosc=ConcurrentReadOutputStream.getStream(ffoutC, null, qfoutC, null, 4, null, false);
			rosc.start(); //Start the stream
		}else{rosc=null;}
		
		//Reset counters
		readsProcessed=0;
		basesProcessed=0;
		
		//Process the read stream
		processInner(cris, ros, rosc);
		
		if(verbose){outstream.println("Finished; closing streams.");}
		
		//Write anything that was accumulated by ReadStats
		errorState|=ReadStats.writeAll();
		//Close the read streams
		errorState|=ReadWrite.closeStreams(cris, ros, rosc);
		
		//Report timing and results
		{
			t.stop();
			
			//Calculate units per nanosecond
			double rpnano=readsProcessed/(double)(t.elapsed);
			double bpnano=basesProcessed/(double)(t.elapsed);
			
			//Add "k" and "m" for large numbers
			String rpstring=(readsProcessed<100000 ? ""+readsProcessed : readsProcessed<100000000 ? (readsProcessed/1000)+"k" : (readsProcessed/1000000)+"m");
			String bpstring=(basesProcessed<100000 ? ""+basesProcessed : basesProcessed<100000000 ? (basesProcessed/1000)+"k" : (basesProcessed/1000000)+"m");
			
			//Format the strings so they have they are right-justified
			while(rpstring.length()<8){rpstring=" "+rpstring;}
			while(bpstring.length()<8){bpstring=" "+bpstring;}
			
			outstream.println("Time:                         \t"+t);
			outstream.println("Reads Processed:    "+rpstring+" \t"+String.format("%.2fk reads/sec", rpnano*1000000));
			outstream.println("Bases Processed:    "+bpstring+" \t"+String.format("%.2fm bases/sec", bpnano*1000));
		}
		
		//Throw an exception of there was an error in a thread
		if(errorState){
			throw new RuntimeException(getClass().getName()+" terminated in an error state; the output may be corrupt.");
		}
	}
	
	/** Iterate through the reads */
	void processInner(final ConcurrentReadInputStream cris, final ConcurrentReadOutputStream ros, final ConcurrentReadOutputStream rosc){
		
		ArrayList<Read> scaffolds=getReads(cris);
		
		final boolean makeLegend=(legendFile!=null);
		TextStreamWriter tswl=(makeLegend ? new TextStreamWriter(legendFile, overwrite, append, false) : null);
		if(makeLegend){tswl.start();}
		
		if(ros!=null){
			if(sortScaffolds){Collections.sort(scaffolds, ReadLengthComparator.comparator);}
			if(renameScaffolds){
				for(Read r : scaffolds){
					String old=r.id;
					r.id="scaffold_"+scafNum;
					if(tswl!=null){tswl.println(old+"\t"+r.id);}
					scafNum++;
				}
			}
			ros.add(scaffolds, 0);
		}
		if(tswl!=null){tswl.poisonAndWait();}

		final boolean makeAgp=(agpFile!=null);
		TextStreamWriter tsw=(makeAgp ? new TextStreamWriter(agpFile, overwrite, append, false) : null);
		if(makeAgp){tsw.start();}
		
		if(rosc!=null){
			ArrayList<Read> contigs=new ArrayList<Read>();
			for(Read r : scaffolds){
				ArrayList<Read> temp=r.breakAtGaps(makeAgp, minContig);
				if(makeAgp){
					tsw.print((String)r.obj);
					r.obj=null;
				}
				contigs.addAll(temp);
			}
			if(sortContigs){Collections.sort(contigs, ReadLengthComparator.comparator);}
			if(renameContigs){
				for(Read r : contigs){
					r.id="contig_"+contigNum;
					contigNum++;
				}
			}
			rosc.add(contigs, 0);
		}
		
		if(makeAgp){tsw.poisonAndWait();}
		
	}
	
	/** Iterate through the reads */
	private ArrayList<Read> getReads(final ConcurrentReadInputStream cris){
		
		ArrayList<Read> all=new ArrayList<Read>(10000);
		
		{
			//Grab the first ListNum of reads
			ListNum<Read> ln=cris.nextList();
			//Grab the actual read list from the ListNum
			ArrayList<Read> reads=(ln!=null ? ln.list : null);
			
			//Check to ensure pairing is as expected
			if(reads!=null && !reads.isEmpty()){
				Read r=reads.get(0);
				assert((ffin1==null || ffin1.samOrBam()) || (r.mate!=null)==cris.paired());
			}
			
			//As long as there is a nonempty read list...
			while(reads!=null && reads.size()>0){
				if(verbose){outstream.println("Fetched "+reads.size()+" reads.");}
				
				//Loop through each read in the list
				for(int idx=0; idx<reads.size(); idx++){
					final Read r1=reads.get(idx);
					assert(r1.mate==null);
					
					//Track the initial length for statistics
					final int initialLength1=r1.length();
					
					//Increment counters
					readsProcessed+=1;
					basesProcessed+=initialLength1;
					
					boolean keep=processRead(r1);
					if(keep){all.add(r1);}
				}
				
				//Notify the input stream that the list was used
				cris.returnList(ln.id, ln.list.isEmpty());
				if(verbose){outstream.println("Returned a list.");}
				
				//Fetch a new list
				ln=cris.nextList();
				reads=(ln!=null ? ln.list : null);
			}
			
			//Notify the input stream that the final list was used
			if(ln!=null){
				cris.returnList(ln.id, ln.list==null || ln.list.isEmpty());
			}
		}
		
		return all;
	}
	
	/*--------------------------------------------------------------*/
	/*----------------         Inner Methods        ----------------*/
	/*--------------------------------------------------------------*/
	
	/**
	 * Process a single read pair.
	 * @param r1 Read 1
	 * @return True if the reads should be kept, false if they should be discarded.
	 */
	boolean processRead(final Read r1){
		assert(!banIupac || !r1.containsNonACGTN()) : "Non-ACGTN base found in scaffold "+r1.id;
		r1.inflateGaps(minGapIn, minGapOut);
		return r1.length()>=minScaf;
	}
	
	/** This is called if the program runs with no parameters */
	private void printOptions(){
		throw new RuntimeException("TODO");
	}
	
	/*--------------------------------------------------------------*/
	/*----------------            Fields            ----------------*/
	/*--------------------------------------------------------------*/
	
	private int minGapIn=1;
	private int minGapOut=10;
	private int minScaf=1;
	private int minContig=1;
	private long scafNum=1;
	private long contigNum=1;
	
	private boolean sortScaffolds=true;
	private boolean sortContigs=false;
	private boolean banIupac=true;
	private boolean renameScaffolds=true;
	private boolean renameContigs=false;
	
	/*--------------------------------------------------------------*/
	/*----------------          I/O Fields          ----------------*/
	/*--------------------------------------------------------------*/

	/** Primary input file path */
	private String in1=null;
	
	private String qfin1=null;

	/** Primary output file path */
	private String out1=null;
	private String outC=null;

	private String qfout1=null;
	private String qfoutC=null;

	private String agpFile=null;
	private String legendFile=null;
	
	/** Override input file extension */
	private String extin=null;
	/** Override output file extension */
	private String extout=null;
	
	/*--------------------------------------------------------------*/

	/** Number of reads processed */
	protected long readsProcessed=0;
	/** Number of bases processed */
	protected long basesProcessed=0;
	
	/** Quit after processing this many input reads; -1 means no limit */
	private long maxReads=-1;
	
	/*--------------------------------------------------------------*/
	/*----------------         Final Fields         ----------------*/
	/*--------------------------------------------------------------*/

	/** Primary input file */
	private final FileFormat ffin1;
	
	/** Primary output file */
	private final FileFormat ffout1;
	private final FileFormat ffoutC;
	
	/*--------------------------------------------------------------*/
	/*----------------        Common Fields         ----------------*/
	/*--------------------------------------------------------------*/
	
	/** Print status messages to this output stream */
	private PrintStream outstream=System.err;
	/** Print verbose messages */
	public static boolean verbose=false;
	/** True if an error was encountered */
	public boolean errorState=false;
	/** Overwrite existing output files */
	private boolean overwrite=false;
	/** Append to existing output files */
	private boolean append=false;
	/** This flag has no effect on singlethreaded programs */
	private final boolean ordered=false;
	
}

==> bbmap-37.62-0.tar.bz2: opt/bbmap-37.62/current/jgi/FungalRelease.java <==
package jgi;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Locale;

import stream.ConcurrentReadInputStream;
import stream.FASTQ;
import stream.FastaReadInputStream;
import stream.ConcurrentReadOutputStream;
import stream.Read;
import structures.ListNum;
import fileIO.ByteFile;
import fileIO.ReadWrite;
import fileIO.FileFormat;
import fileIO.TextStreamWriter;
import shared.Parser;
import shared.ReadStats;
import shared.Shared;
import shared.Timer;
import shared.Tools;
import sort.ReadLengthComparator;

/**
 * Reformats a fungal assembly for release.
 * Also creates contig and agp files.
 * 
 * @author Brian Bushnell
 * @date December 9, 2015
 *
 */
public class FungalRelease {
	
	/*--------------------------------------------------------------*/
	/*----------------        Initialization        ----------------*/
	/*--------------------------------------------------------------*/
	
	/**
	 * Code entrance from the command line.
	 * @param args Command line arguments
	 */
	public static void main(String[] args){
		Timer t=new Timer();
		FungalRelease fr=new FungalRelease(args);
		fr.process(t);
	}
	
	/**
	 * Constructor.
	 * @param args Command line arguments
	 */
	public FungalRelease(String[] args){
		
		//Process any config files
		args=Parser.parseConfig(args);
		
		//Detect whether the uses needs help
		if(Parser.parseHelp(args, true)){
			printOptions();
			System.exit(0);
		}
		
		//Print the program name and arguments
		outstream.println("Executing "+getClass().getName()+" "+Arrays.toString(args)+"\n");
		
		FASTQ.FORCE_INTERLEAVED=FASTQ.TEST_INTERLEAVED=false;
		Shared.FASTA_WRAP=60;
		
		//Set some shared static variables regarding PIGZ
		Shared.capBufferLen(200);
		Shared.capBuffers(4);
		ReadWrite.USE_PIGZ=ReadWrite.USE_UNPIGZ=true;
		ReadWrite.MAX_ZIP_THREADS=Shared.threads();
		
		Read.TO_UPPER_CASE=true;
		
		//Create a parser object
		Parser parser=new Parser();
		
		//Parse each argument
		for(int i=0; i<args.length; i++){
			String arg=args[i];
			
			//Break arguments into their constituent parts, in the form of "a=b"
			String[] split=arg.split("=");
			String a=split[0].toLowerCase();
			String b=split.length>1 ? split[1] : null;
			if(b==null || b.equalsIgnoreCase("null")){b=null;}
			while(a.startsWith("-")){a=a.substring(1);} //Strip leading hyphens
			
			
			if(parser.parse(arg, a, b)){//Parse standard flags in the parser
				//do nothing
			}else if(a.equals("verbose")){
				verbose=Tools.parseBoolean(b);
			}else if(a.equals("mingapin")){
				minGapIn=(int)Tools.parseKMG(b);
			}else if(a.equals("mingap") || a.equals("mingapout")){
				minGapOut=(int)Tools.parseKMG(b);
			}else if(a.equals("minlen") || a.equals("minlength") || a.equals("minscaf")){
				minScaf=(int)Tools.parseKMG(b);
			}else if(a.equals("mincontig")){
				minContig=(int)Tools.parseKMG(b);
			}else if(a.equals("outc") || a.equals("contigs")){
				outC=b;
			}else if(a.equals("qfoutc")){
				qfoutC=b;
			}else if(a.equals("sortcontigs")){
				sortContigs=Tools.parseBoolean(b);
			}else if(a.equals("sortcscaffolds")){
				sortScaffolds=Tools.parseBoolean(b);
			}else if(a.equals("baniupac")){
				banIupac=Tools.parseBoolean(b);
			}else if(a.equals("agp")){
				agpFile=b;
			}else if(a.equals("legend")){
				legendFile=b;
			}else if(a.equals("scafnum")){
				scafNum=Tools.parseKMG(b);
			}else if(a.equals("renamescaffolds") || a.equals("rename")){
				renameScaffolds=Tools.parseBoolean(b);
			}else if(a.equals("scafnum")){
				contigNum=Tools.parseKMG(b);
			}else if(a.equals("renamecontigs")){
				renameContigs=Tools.parseBoolean(b);
			}else if(a.equals("parse_flag_goes_here")){
				//Set a variable here
			}else{
				outstream.println("Unknown parameter "+args[i]);
				assert(false) : "Unknown parameter "+args[i];
				//				throw new RuntimeException("Unknown parameter "+args[i]);
			}
		}
		
		{//Process parser fields
			Parser.processQuality();
			
			maxReads=parser.maxReads;
			
			overwrite=ReadStats.overwrite=parser.overwrite;
			append=ReadStats.append=parser.append;
			
			in1=parser.in1;
			qfin1=parser.qfin1;
			
			out1=parser.out1;
			qfout1=parser.qfout1;
			
			extin=parser.extin;
			extout=parser.extout;
		}
		
		assert(FastaReadInputStream.settingsOK());
		
		//Ensure there is an input file
		if(in1==null){
			printOptions();
			throw new RuntimeException("Error - at least one input file is required.");
		}
		
		//Adjust the number of threads for input file reading
		if(!ByteFile.FORCE_MODE_BF1 && !ByteFile.FORCE_MODE_BF2 && Shared.threads()>2){
			ByteFile.FORCE_MODE_BF2=true;
		}
		
		//Ensure output files can be written
		if(!Tools.testOutputFiles(overwrite, append, false, out1, outC)){
			outstream.println((out1==null)+", "+out1);
			throw new RuntimeException("\n\noverwrite="+overwrite+"; Can't write to output files "+out1+"\n");
		}
		
		//Ensure input files can be read
		if(!Tools.testInputFiles(false, true, in1)){
			throw new RuntimeException("\nCan't read to some input files.\n");
		}
		
		//Ensure that no file was specified multiple times
		if(!Tools.testForDuplicateFiles(true, in1, out1, outC)){
			throw new RuntimeException("\nSome file names were specified multiple times.\n");
		}
		
		//Create output FileFormat objects
		ffout1=FileFormat.testOutput(out1, FileFormat.FASTA, extout, true, overwrite, append, ordered);
		
		//Create output FileFormat objects
		ffoutC=FileFormat.testOutput(outC, FileFormat.FASTA, extout, true, overwrite, append, ordered);

		//Create input FileFormat objects
		ffin1=FileFormat.testInput(in1, FileFormat.FASTA, extin, true, true);
	}
	
	/*--------------------------------------------------------------*/
	/*----------------         Outer Methods        ----------------*/
	/*--------------------------------------------------------------*/

	/** Create read streams and process all data */
	void process(Timer t){
		
		//Create a read input stream
		final ConcurrentReadInputStream cris;
		{
			cris=ConcurrentReadInputStream.getReadInputStream(maxReads, true, ffin1, null, qfin1, null);
			cris.start(); //Start the stream
			if(verbose){outstream.println("Started cris");}
		}
		
		//Optionally create a read output stream
		final ConcurrentReadOutputStream ros, rosc;
		final int buff=4;
		if(ffout1!=null){
			ros=ConcurrentReadOutputStream.getStream(ffout1, null, qfout1, null, buff, null, false);
			ros.start(); //Start the stream
		}else{ros=null;}
		if(ffoutC!=null){
			rosc=ConcurrentReadOutputStream.getStream(ffoutC, null, qfoutC, null, 4, null, false);
			rosc.start(); //Start the stream
		}else{rosc=null;}
		
		//Reset counters
		readsProcessed=0;
		basesProcessed=0;
		
		//Process the read stream
		processInner(cris, ros, rosc);
		
		if(verbose){outstream.println("Finished; closing streams.");}
		
		//Write anything that was accumulated by ReadStats
		errorState|=ReadStats.writeAll();
		//Close the read streams
		errorState|=ReadWrite.closeStreams(cris, ros, rosc);
		
		//Report timing and results
		{
			t.stop();
			
			//Calculate units per nanosecond
			double rpnano=readsProcessed/(double)(t.elapsed);
			double bpnano=basesProcessed/(double)(t.elapsed);
			
			//Add "k" and "m" for large numbers
			String rpstring=(readsProcessed<100000 ? ""+readsProcessed : readsProcessed<100000000 ? (readsProcessed/1000)+"k" : (readsProcessed/1000000)+"m");
			String bpstring=(basesProcessed<100000 ? ""+basesProcessed : basesProcessed<100000000 ? (basesProcessed/1000)+"k" : (basesProcessed/1000000)+"m");
			
			//Format the strings so they have they are right-justified
			while(rpstring.length()<8){rpstring=" "+rpstring;}
			while(bpstring.length()<8){bpstring=" "+bpstring;}
			
			outstream.println("Time:                         \t"+t);
			outstream.println("Reads Processed:    "+rpstring+" \t"+String.format(Locale.ROOT, "%.2fk reads/sec", rpnano*1000000));
			outstream.println("Bases Processed:    "+bpstring+" \t"+String.format(Locale.ROOT, "%.2fm bases/sec", bpnano*1000));
		}
		
		//Throw an exception of there was an error in a thread
		if(errorState){
			throw new RuntimeException(getClass().getName()+" terminated in an error state; the output may be corrupt.");
		}
	}
	
	/** Iterate through the reads */
	void processInner(final ConcurrentReadInputStream cris, final ConcurrentReadOutputStream ros, final ConcurrentReadOutputStream rosc){
		
		ArrayList<Read> scaffolds=getReads(cris);
		
		final boolean makeLegend=(legendFile!=null);
		TextStreamWriter tswl=(makeLegend ? new TextStreamWriter(legendFile, overwrite, append, false) : null);
		if(makeLegend){tswl.start();}
		
		if(ros!=null){
			if(sortScaffolds){Shared.sort(scaffolds, ReadLengthComparator.comparator);}
			if(renameScaffolds){
				for(Read r : scaffolds){
					String old=r.id;
					r.id="scaffold_"+scafNum;
					if(tswl!=null){tswl.println(old+"\t"+r.id);}
					scafNum++;
				}
			}
			ros.add(scaffolds, 0);
		}
		if(tswl!=null){tswl.poisonAndWait();}

		final boolean makeAgp=(agpFile!=null);
		TextStreamWriter tsw=(makeAgp ? new TextStreamWriter(agpFile, overwrite, append, false) : null);
		if(makeAgp){tsw.start();}
		
		if(rosc!=null || makeAgp){//Process contigs
			ArrayList<Read> contigs=new ArrayList<Read>();
			for(Read r : scaffolds){
				ArrayList<Read> temp=r.breakAtGaps(makeAgp, minContig);
				if(makeAgp){
					tsw.print((String)r.obj);
					r.obj=null;
				}
				contigs.addAll(temp);
			}
			if(sortContigs){Shared.sort(contigs, ReadLengthComparator.comparator);}
			if(renameContigs){
				for(Read r : contigs){
					r.id="contig_"+contigNum;
					contigNum++;
				}
			}
			if(rosc!=null){
				rosc.add(contigs, 0);
			}
		}
		
		if(makeAgp){tsw.poisonAndWait();}
		
	}
	
	/** Iterate through the reads */
	private ArrayList<Read> getReads(final ConcurrentReadInputStream cris){
		
		ArrayList<Read> all=new ArrayList<Read>(10000);
		
		{
			//Grab the first ListNum of reads
			ListNum<Read> ln=cris.nextList();
			//Grab the actual read list from the ListNum
			ArrayList<Read> reads=(ln!=null ? ln.list : null);
			
			//Check to ensure pairing is as expected
			if(reads!=null && !reads.isEmpty()){
				Read r=reads.get(0);
				assert((ffin1==null || ffin1.samOrBam()) || (r.mate!=null)==cris.paired());
			}
			
			//As long as there is a nonempty read list...
			while(reads!=null && reads.size()>0){
				if(verbose){outstream.println("Fetched "+reads.size()+" reads.");}
				
				//Loop through each read in the list
				for(int idx=0; idx<reads.size(); idx++){
					final Read r1=reads.get(idx);
					assert(r1.mate==null);
					
					//Track the initial length for statistics
					final int initialLength1=r1.length();
					
					//Increment counters
					readsProcessed+=1;
					basesProcessed+=initialLength1;
					
					boolean keep=processRead(r1);
					if(keep){all.add(r1);}
				}
				
				//Notify the input stream that the list was used
				cris.returnList(ln.id, ln.list.isEmpty());
				if(verbose){outstream.println("Returned a list.");}
				
				//Fetch a new list
				ln=cris.nextList();
				reads=(ln!=null ? ln.list : null);
			}
			
			//Notify the input stream that the final list was used
			if(ln!=null){
				cris.returnList(ln.id, ln.list==null || ln.list.isEmpty());
			}
		}
		
		return all;
	}
	
	/*--------------------------------------------------------------*/
	/*----------------         Inner Methods        ----------------*/
	/*--------------------------------------------------------------*/
	
	/**
	 * Process a single read pair.
	 * @param r1 Read 1
	 * @return True if the reads should be kept, false if they should be discarded.
	 */
	boolean processRead(final Read r1){
		assert(!banIupac || !r1.containsNonACGTN()) : "Non-ACGTN base found in scaffold "+r1.id;
		r1.inflateGaps(minGapIn, minGapOut);
		return r1.length()>=minScaf;
	}
	
	/** This is called if the program runs with no parameters */
	private void printOptions(){assert(false) : "Please read the associated shell script for usage information.";}
	
	/*--------------------------------------------------------------*/
	/*----------------            Fields            ----------------*/
	/*--------------------------------------------------------------*/
	
	private int minGapIn=1;
	private int minGapOut=10;
	private int minScaf=1;
	private int minContig=1;
	private long scafNum=1;
	private long contigNum=1;
	
	private boolean sortScaffolds=true;
	private boolean sortContigs=false;
	private boolean banIupac=true;
	private boolean renameScaffolds=true;
	private boolean renameContigs=false;
	
	/*--------------------------------------------------------------*/
	/*----------------          I/O Fields          ----------------*/
	/*--------------------------------------------------------------*/

	/** Primary input file path */
	private String in1=null;
	
	private String qfin1=null;

	/** Primary output file path */
	private String out1=null;
	private String outC=null;

	private String qfout1=null;
	private String qfoutC=null;

	private String agpFile=null;
	private String legendFile=null;
	
	/** Override input file extension */
	private String extin=null;
	/** Override output file extension */
	private String extout=null;
	
	/*--------------------------------------------------------------*/

	/** Number of reads processed */
	protected long readsProcessed=0;
	/** Number of bases processed */
	protected long basesProcessed=0;
	
	/** Quit after processing this many input reads; -1 means no limit */
	private long maxReads=-1;
	
	/*--------------------------------------------------------------*/
	/*----------------         Final Fields         ----------------*/
	/*--------------------------------------------------------------*/

	/** Primary input file */
	private final FileFormat ffin1;
	
	/** Primary output file */
	private final FileFormat ffout1;
	private final FileFormat ffoutC;
	
	/*--------------------------------------------------------------*/
	/*----------------        Common Fields         ----------------*/
	/*--------------------------------------------------------------*/
	
	/** Print status messages to this output stream */
	private PrintStream outstream=System.err;
	/** Print verbose messages */
	public static boolean verbose=false;
	/** True if an error was encountered */
	public boolean errorState=false;
	/** Overwrite existing output files */
	private boolean overwrite=false;
	/** Append to existing output files */
	private boolean append=false;
	/** This flag has no effect on singlethreaded programs */
	private final boolean ordered=false;
	
}
